using System;
using System.Collections.Generic;
using System.Text;
using EGIS.ShapeFileLib;
using System.Drawing;
using System.Drawing.Imaging;
using System.Net.Cache;
using System.Net;
using System.Threading;

/*
 * 
 * DISCLAIMER OF WARRANTY: THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
 * WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE OR NON-INFRINGING.
 * THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH YOU. SHOULD ANY COVERED CODE PROVE DEFECTIVE IN ANY RESPECT,
 * YOU (NOT CORPORATE EASY GIS .NET) ASSUME THE COST OF ANY NECESSARY SERVICING, REPAIR OR CORRECTION.  
 * 
 * LIABILITY: IN NO EVENT SHALL CORPORATE EASY GIS .NET BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING, WITHOUT LIMITATION, 
 * DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF INFORMATION OR ANY OTHER PECUNIARY LOSS)
 * ARISING OUT OF THE USE OF INABILITY TO USE THIS SOFTWARE, EVEN IF CORPORATE EASY GIS .NET HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGES.
 * 
 * Copyright: Winston Fletcher 2010 - 2015
 *
 */

namespace Example6
{
    class TileCollection
    {
        private int tileCountX;
        private int tileCountY;
        Tile[] tiles;

        public TileCollection(double mapScale, PointD mapCenter, int pixelWidth, int pixelHeight, EGIS.Controls.SFMap map,
            string imageUrlFormat)
        {
            int zoomLevel = TileUtil.ScaleToZoomLevel(mapScale);
            this.tileCountX = (int)Math.Ceiling(pixelWidth / 256.0) + 1;
            this.tileCountY = (int)Math.Ceiling(pixelHeight / 256.0) + 1;
            tiles = new Tile[tileCountX * tileCountY];
            PointD mapOffset = GetTopLeftGisPoint(pixelWidth, pixelHeight, mapCenter, mapScale);
            mapOffset = ShapeFile.MercatorToLL(GetTopLeftGisPoint(pixelWidth, pixelHeight, ShapeFile.LLToMercator(mapCenter), mapScale));
            System.Drawing.Point tileOffset = TileUtil.GetTileFromGisLocation(mapOffset.X, mapOffset.Y, zoomLevel);
            long mapPixOffX, mapPixOffY;
            TileUtil.LLToPixel(mapOffset, zoomLevel, out mapPixOffX, out mapPixOffY);
            mapPixOffX = (tileOffset.X * 256) - mapPixOffX;
            mapPixOffY = (tileOffset.Y * 256) - mapPixOffY;
            int dy = (int)mapPixOffY;
            int tileIndex = 0;
            for (int y = 0; y < tileCountY; ++y)
            {
                int dx = (int)mapPixOffX;
                for (int x = 0; x < tileCountX; ++x)
                {
                    tiles[tileIndex] = new Tile(tileOffset.X + x, tileOffset.Y + y, zoomLevel, dx, dy, map, imageUrlFormat);
                    dx += 256;
                    tileIndex++;
                }
                dy += 256;
            }
        }

        public TileCollection(double mapScale, PointD mapCenter, int pixelWidth, int pixelHeight, EGIS.Controls.SFMap map)
            : this(mapScale, mapCenter, pixelWidth, pixelHeight, map, Tile.DefaultImageUrl)
        {
                        
        }

        private static PointD GetTopLeftGisPoint(int pixWidth, int pixHeight, PointD mapCenter, double mapScale)
        {            
            double s = 1.0 / mapScale;
            return new PointD(mapCenter.X - (s * ((pixWidth >> 1) - 0)), mapCenter.Y + (s * ((pixHeight >> 1) - 0)));
        }

    
        /// <summary>
        /// Render the TileCollection
        /// </summary>
        /// <param name="g"></param>
        /// <param name="transparency">trapsarency between 0.0 - 1.0 </param>
        public void Render(Graphics g, float transparency)
        {
            for (int n = 0; n < this.tiles.Length; ++n)
            {
                tiles[n].Render(g, transparency);
            }                   
        }        
    }

    
    class Tile
    {        
        private EGIS.Controls.SFMap map;
        private int x, y;
        private int zoomLevel;
        private int pixOffX, pixOffY;
        private System.Net.WebRequest imageWebRequest;
        private string imageUrlFormat = "";

        internal const string DefaultImageUrl = "http://b.tile.openstreetmap.org/{0}/{1}/{2}.png";

        public Tile(int x, int y, int zoomLevel, int pixOffX, int pixOffY, EGIS.Controls.SFMap map)
            : this(x, y, zoomLevel, pixOffX, pixOffY, map, DefaultImageUrl)
        {
            
        }

        public Tile(int x, int y, int zoomLevel, int pixOffX, int pixOffY, EGIS.Controls.SFMap map, string imageUrlFormat)
        {
            this.x = x;
            this.y = y;
            this.zoomLevel = zoomLevel;
            this.pixOffX = pixOffX;
            this.pixOffY = pixOffY;
            this.map = map;
            this.imageUrlFormat = imageUrlFormat;
        }

               
        private const HttpRequestCacheLevel CacheLevel = HttpRequestCacheLevel.CacheIfAvailable;


        #region methods to handle pending requests
        
        private static System.Collections.Concurrent.ConcurrentDictionary<string, bool> pendingRequests = new System.Collections.Concurrent.ConcurrentDictionary<string, bool>();

        private static bool RequestPending(string url)
        {
            return pendingRequests.ContainsKey(url);
        }

        private static void RemovePendingRequest(string url)
        {
            bool b;
            pendingRequests.TryRemove(url, out b);
        }

        private static void AddPendingRequest(string url)
        {
            pendingRequests[url] = true;
        }

        private static void RespCallback(IAsyncResult asynchronousResult)
        {
            // Set the State of request to asynchronous.
            Tile tile = (Tile)asynchronousResult.AsyncState;
            try
            {
                // End the Asynchronous response.              
                WebResponse response = tile.imageWebRequest.EndGetResponse(asynchronousResult);
                System.IO.Stream responseStream = response.GetResponseStream();
                using (Image img = Bitmap.FromStream(responseStream))
                {
                    response.Close();
                    responseStream.Close();
                }
                //tile.map.Refresh(true); //can't call Refresh as functin is not called on main thread
                tile.map.InvalidateAndClearBackground();
            }
            catch { }
            //catch (WebException)
            //{
            //}
            //catch (Exception)
            //{
            //}
            finally
            {
                RemovePendingRequest(tile.imageWebRequest.RequestUri.OriginalString);
            }
        }
       
        #endregion


        private void GetBitmapAsync()
        {
            string strUrl = string.Format(imageUrlFormat, this.zoomLevel, this.x, this.y);
            if (RequestPending(strUrl))
            {
                return; //request already pending
            }
            
            imageWebRequest = System.Net.WebRequest.Create(strUrl);
            HttpWebRequest hwr = imageWebRequest as HttpWebRequest;
            if (hwr != null)
            {
                if (hwr.ServicePoint.ConnectionLimit < 4)
                {
                    hwr.ServicePoint.ConnectionLimit = 4;
                }
            }
            imageWebRequest.CachePolicy = new HttpRequestCachePolicy(CacheLevel);
            AddPendingRequest(strUrl);
            IAsyncResult asyncResult = (IAsyncResult)imageWebRequest.BeginGetResponse(new AsyncCallback(RespCallback), this);
        }

        
        private System.Drawing.Bitmap CreateBitmap()
        {
            //http://tah.openstreetmap.org/Tiles/tile/0/0/0.png
            //string strUrl = string.Format("http://tah.openstreetmap.org/Tiles/tile/{0}/{1}/{2}.png", this.zoomLevel, this.x, this.y);
            string strUrl = string.Format(imageUrlFormat, this.zoomLevel, this.x, this.y);
            System.Net.WebRequest r = System.Net.WebRequest.Create(strUrl);
            //attempt to retrieve from the local cache
            r.CachePolicy = new HttpRequestCachePolicy(HttpRequestCacheLevel.CacheOnly);

            try
            {
                System.Net.WebResponse response = r.GetResponse();
                System.IO.Stream responseStream = response.GetResponseStream();
                using (Image img = Bitmap.FromStream(responseStream))
                {
                    response.Close();
                    responseStream.Close();

                    //copy the returned image to a new bitmap. If we don't do this then the transparency may not
                    //display properly
                    Bitmap bm = new Bitmap(img.Width, img.Height);
                    using (Graphics g = Graphics.FromImage(bm))
                    {
                        g.DrawImage(img, new Rectangle(0,0,img.Width, img.Height),0,0,img.Width,img.Height,GraphicsUnit.Pixel);
                    }

                    return bm;
                }
            }
            catch (WebException wex)
            {
                if (wex.Status == WebExceptionStatus.CacheEntryNotFound)
                {
                    //request not in local cache - retrieve asyncronously
                    GetBitmapAsync();
                }
                Bitmap blankBitmap = new Bitmap(256, 256);
                using (Graphics g = Graphics.FromImage(blankBitmap))
                {
                    g.Clear(Color.LightGray);
                }
                return blankBitmap;                
            }            
        }

               
        internal void Render(Graphics g, float transparency)
        {
            using(Image bm = this.CreateBitmap())
            {
                float[][] ptsArray ={ 
                    new float[] {1, 0, 0, 0, 0},
                    new float[] {0, 1, 0, 0, 0},
                    new float[] {0, 0, 1, 0, 0},
                    new float[] {0, 0, 0, transparency, 0}, 
                    new float[] {0, 0, 0, 0, 1}};
                ColorMatrix clrMatrix = new ColorMatrix(ptsArray);
                ImageAttributes imgAttributes = new ImageAttributes();
                imgAttributes.SetColorMatrix(clrMatrix, ColorMatrixFlag.Default, ColorAdjustType.Bitmap);
                g.DrawImage(bm, new Rectangle(pixOffX, pixOffY, bm.Width, bm.Height), 0, 0, bm.Width, bm.Height, GraphicsUnit.Pixel, imgAttributes);
            }
        }   
 
    }
}
